/* lzoLib.c : [De]Compression using LZO
 * Bog simple, very fast compression and decompression
 * http://www.oberhumer.com/opensource/lzo/
 * 
 * Compile multi-threaded (MSVC properties/C/Code Generation/Runtime library)
 * Linked against zkl.lib (zkl exports)
 *    msvcrt.lib.
 *    In MSVC properties/Linker/Input, set "Ignore All Default Libraries" to
 *    "Yes (/NODEFAULTLIB)". This creates a much smaller DLL and is why you
 *    need to link against msvcrt.lib
 * Creates: lzoLib.lib (exports), lzoLib.dll
 * 
 * Thread safe
 * 
 * Copyright (c) 2006, 2007, 2008 Craig Durland craigd@zenkinetic.com
 * License: The zlib/libpng License,
 * 	http://www.opensource.org/licenses/zlib-license.php
 * 
 * The LZO code is covered by the GPL, see minilzo-202/COPYING, which may
 * supercede my license. Ask somebody who knows if it matters.
 */

#define _CRT_SECURE_NO_WARNINGS	// don't bitch about stuff like sprintf

#include <stdio.h>

#include "zklObject.h"	
#include "zklAtomic.h"
#include "zklData.h"
#include "zklList.h"
#include "zklNumber.h"
#include "zklString.h"
#include "zklUtil.h"
#include "zklImports.h"

#if defined( __linux__)	// package liblzo2-dev
   #include <minilzo/minilzo.h>
#else
   #include "minilzo-202/minilzo.h"
#endif

static ZKL_Object LZOObject;
static SpinLock   lock;

#define OUT_LEN(inLen)	(inLen + inLen / 16 + 64 + 3)

    /* Work-memory needed for compression.  Allocate memory in units of
     * `lzo_align_t' (instead of `char') to make sure it is properly
     * aligned.
     */

#define HEAP_ALLOC(var,size) \
    lzo_align_t __LZO_MMODEL var \
        [ ((size) + (sizeof(lzo_align_t) - 1)) / sizeof(lzo_align_t) ]

    	// LZO1X_1_MEM_COMPRESS == 65536 on Win32
static HEAP_ALLOC(wrkmem,LZO1X_1_MEM_COMPRESS);	// ick, global shared mem

/* ******************************************************************** */
/* ***************************** Methods ****************************** */
/* ******************************************************************** */

    // .compress(data) --> L(srcLen, checksum, compressed data)
static Instance *compress(Instance *self,pArglist arglist,pVM vm)
{
   Instance    *data = arglistGetBT(arglist,0,DataType,"compress",vm);
   int		r;
   lzo_uint	srcLen, dstLen;
   lzo_uint32	compressedChecksum;
   size_t	dataSize;
   Instance    *output;
   Byte	       *src, *dst;

   src	  = dataText(data,&dataSize);
   output = dataCreate(OUT_LEN(dataSize),I_OWNED,vm);
   dst    = dataWillFill(output);

	// Step 1: initialize the LZO library
   if (lzo_init() != LZO_E_OK) vmThrow(vm,E_IO_ERROR,"lzo_init() failed");

	// Compress with LZO1X-1
   srcLen = (lzo_uint)dataSize;
   SPIN_LOCK_ACQUIRE(&lock);
      r = lzo1x_1_compress(src,srcLen,dst,&dstLen,wrkmem);
   SPIN_LOCK_RELEASE(&lock);

   if (r == LZO_E_OK) dataSetSize(output,dstLen);
   else	// this should NEVER happen.
   {
      char	buf[200];
      sprintf(buf,"Compression failed: %d",r);
      vmThrow(vm,E_IO_ERROR,buf);
   }

	// check for an incompressible block
   if (dstLen >= srcLen) return BoolFalse;

   compressedChecksum = lzo_adler32(0,NULL,0);
   compressedChecksum = lzo_adler32(compressedChecksum,dst,dstLen);

   return tupleCreateX(vm,
	intCreate(srcLen,vm), intCreate(compressedChecksum,vm),
	output,ZNIL);
}

    // .decompress(length, checksum, compressed data) --> Data
static Instance *decompress(Instance *self,pArglist arglist,pVM vm)
{
   lzo_uint	dstLen   = (lzo_uint)arglistGetInt(arglist,0,0,vm);
   lzo_uint32	checkSum = (lzo_uint32)arglistGetInt(arglist,1,0,vm);
   Instance    *data     = arglistGetBT(arglist,2,DataType,"decompress",vm);
   size_t	dataSize;
   int		r;
   Instance    *output;
   Byte	       *src, *dst;
   lzo_uint	srcLen, outLen;
   lzo_uint32	dataCheckSum;

   src = dataText(data,&dataSize); srcLen = (lzo_uint)dataSize;

   dataCheckSum = lzo_adler32(0,NULL,0);
   dataCheckSum = lzo_adler32(dataCheckSum,src,srcLen);

   if (dataCheckSum != checkSum)
      vmThrow(vm,E_VALUE_ERROR,"decompress: checksum mismatch");

   if (lzo_init() != LZO_E_OK) vmThrow(vm,E_IO_ERROR,"lzo_init() failed");

   output = dataCreate(dstLen,I_OWNED,vm);
   dst    = dataWillFill(output);

   outLen = dstLen;
   SPIN_LOCK_ACQUIRE(&lock);
      r = lzo1x_decompress_safe(src,srcLen,dst,&outLen,0);
//      r = lzo1x_decompress(src,srcLen,dst,&outLen,0);
   SPIN_LOCK_RELEASE(&lock);
   if (r == LZO_E_OK)
   {
      if (outLen != dstLen) vmThrow(vm,E_VALUE_ERROR,"decompress: size diff");
      dataSetSize(output,outLen);
   }
   else // this should NEVER happen. Well, on over runs it does
   {
      char buf[200];
      sprintf(buf,"Deompression failed: %d",r);
      vmThrow(vm,E_IO_ERROR,buf);
   }
   return output;
}

    // .calcAdler32(Data)
static Instance *calcAdler32(Instance *self,pArglist arglist,pVM vm)
{
   Instance    *data = arglistGetBT(arglist,0,DataType,"calcAdler32",vm);
   lzo_uint32	checksum;
   size_t	srcSize;
   Byte	       *src = dataText(data,&srcSize);

   checksum = lzo_adler32(0,NULL,0);
   checksum = lzo_adler32(checksum,src,(lzo_uint)srcSize);

   return intCreate(checksum,vm);
}

static MethodTable methods[] =
{
   "compress",		(pMethod)compress,
   "decompress",	(pMethod)decompress,
   "adler32",		(pMethod)calcAdler32,
   0,			0
};


/* ******************************************************************** */
/* **************************** Properties **************************** */
/* ******************************************************************** */

    // .version
static Instance *version(Instance *self, pVM vm)
   { return intCreate(lzo_version(),vm); }	// 0x2020

    // .versionString
static Instance *versionString(Instance *self, pVM vm)
   { return stringCreate((char *)lzo_version_string(),I_OWNED,vm); }

    // .versionDate
static Instance *versionDate(Instance *self, pVM vm)
   { return stringCreate((char *)lzo_version_date(),I_OWNED,vm); }

    // .authors
static Instance *authors(Instance *self, pVM vm)
{
   return createTupleOfStrings(vm,"Craig Durland","http://zenkinetic.com/",
          (char *)0); 
}

    // *.libraries
static Instance *libraries(Instance *self, pVM vm)
{
   return createTupleOfStrings(vm,
      "LZO library: Markus Franz Xaver Johannes Oberhumer",
         "http://www.oberhumer.com/opensource/lzo/",
      (char *)0);
}

static PropertyTable properties[] =
{
   "version",		(pProperty)version,
   "versionString",	(pProperty)versionString,
   "versionDate",	(pProperty)versionDate,
   "authors",		(pProperty)authors,
   "libraries",		(pProperty)libraries,
   0,			0
};



Instance lzoInstance;

DllExport void *construct(void *vm)
{
   if (!vm) return (void *)ZKLX_PROTOCOL;	// handshake
   	// If this is reloaded, nothing happens except 
	// construct() is called again so don't reinitialize
   if (!LZOObject.name)	// static items are zero
   {
      constructObject(&LZOObject,NativeType, methods,properties,0,vm);
      LZOObject.name	   = "LZO";
      LZOObject.vaultPath  = "Utils.Compression";
      LZOObject.isize      = sizeof(lzoInstance);
      LZOObject.threadSafe = 1;

      instanceInit(&lzoInstance,&LZOObject,I_UNTOUCHABLE);

      spinLockInit(&lock);
   }
   return &lzoInstance;
}
